module CoolGirlChina # (
		parameter DOUBLE_WRITE_FIX = 0,				// block two writes in a row (RMW) for games like Snow Bros. and Bill & Ted's Excellent Adventure
		parameter USE_MAPPER_001 = 1,					// mapper #001 - MMC1
		parameter USE_MAPPER_002 = 1,					// mapper #002 - UxROM
		parameter USE_MAPPER_003 = 1,					// mapper #003 - CNROM
		parameter USE_MAPPER_004 = 0,					// mapper #004 - MMC3/MMC6
		parameter USE_MAPPER_007 = 1,					// mapper #007 - AxROM
		parameter USE_MAPPER_009_010 = 0,			// mapper #009 - MMC2, #10 - MMC4
		parameter USE_MAPPER_011 = 1,					// mapper #011 - Color Dreams
		parameter USE_MAPPER_021_022_023_025 = 0,	// mappers #021, #022, #023, #025 - VRC2, VRC4
		parameter USE_MAPPER_022 = 1,					// mapper #022 - VRC2a (shifted CHR lines)
		parameter USE_VRC4_INTERRUPTS = 1,			// for VRC4
		parameter USE_MAPPER_033_048 = 0,			// mappers #033 & #048 - Taito
		parameter USE_MAPPER_048_INTERRUPTS = 1,	// mapper #048 - Taito
		parameter USE_MAPPER_066 = 1,					// mapper #066 - GxROM
		parameter USE_MAPPER_069 = 0, 				// mapper #069 - Sunsoft
		parameter USE_MAPPER_071 = 0,					// mapper #071 (for Fire Hawk only), require USE_MAPPER_002
		parameter USE_MAPPER_073 = 1,					// mapper #073 - VRC3
		parameter USE_MAPPER_090 = 0,					// mapper #090 - JY, for Aladdin only
		parameter USE_MAPPER_118 = 0,					// mapper #118 - TxSROM, require MAPPER_004
		parameter USE_MAPPER_152 = 1,					// mapper #152
		parameter USE_MAPPER_189 = 0,					// mapper #189   require MAPPER_004
		parameter USE_MAPPER_241 = 1, 				// mapper #241 - BNROM, require USE_MAPPER_007

		parameter MMC3_BITSIZE = 8						// Maximum size for MMC3 PRG (6=512KB - standard size, 8=2MB - required for some hacks)
)
(
	input m2,
	input romsel,
	input cpu_rw_in,
	input [3:0] cpu_addr_in_low,
	input [14:8] cpu_addr_in,
	inout [7:0] cpu_data_in,
	output [26:13] cpu_addr_out,
	output flash_we,
	output flash_oe,
	output sram_ce,
	output sram_we,
	output sram_oe,

	input ppu_rd_in,
	input [13:0] ppu_addr_in,
	output [17:10] ppu_addr_out,
	output ppu_ciram_a10,

	output irq
);
	assign cpu_addr_out[26:13] = {cpu_base[26:14] | (cpu_addr_mapped[20:14] & ~prg_mask[20:14]), cpu_addr_mapped[13]};
	assign ppu_addr_out[17:10] = {ppu_addr_mapped[17:13] & ~chr_mask[17:13], ppu_addr_mapped[12:10]};

	assign cpu_data_in = cpu_data_out_enabled ? cpu_data_out : 8'bZZZZZZZZ;
	wire flash_ce_w = romsel;
	assign flash_oe = ~cpu_rw_in | flash_ce_w;

	assign flash_we = cpu_rw_in | flash_ce_w | ~prg_write_enabled;
	wire sram_ce_w = ~(cpu_addr_in[14] & cpu_addr_in[13] & m2 & romsel & sram_enabled);
	assign sram_ce = sram_ce_w;
	assign sram_we = cpu_rw_in | sram_ce_w;
	assign sram_oe = ~cpu_rw_in | sram_ce_w;

	// watchdog
	reg [2:0] wdt_counter = 0;
	reg reset = 0;

	// gated write clock
	reg cpu_rw_in_store = 0;
	reg write_clock = 0;

	reg [26:14] cpu_base = 0;
	reg [20:14] prg_mask = 7'b1111000;
	reg [17:13] chr_mask = 0;
	reg [2:0] prg_mode = 0;
	reg [7:0] prg_bank_a = 0;
	reg [7:0] prg_bank_b = 0; //1;
	reg [7:0] prg_bank_c = 8'b11111110;
	reg [7:0] prg_bank_d = 8'b11111111;
	reg [2:0] chr_mode = 0;
	reg [7:0] chr_bank_a = 0;
	reg [7:0] chr_bank_b = 0; //1;
	reg [7:0] chr_bank_c = 0; //2;
	reg [7:0] chr_bank_d = 0; //3;
	reg [7:0] chr_bank_e = 0; //4;
	reg [7:0] chr_bank_f = 0; //5;
	reg [7:0] chr_bank_g = 0; //6;
	reg [7:0] chr_bank_h = 0; //7;
	reg [4:0] mapper = 0;
	reg [2:0] flags = 0;
	reg sram_enabled = 0;
	reg prg_write_enabled = 0;
	reg [1:0] mirroring = 0;
	reg lockout = 0;

	// some common registers for all mappers
	reg [7:0] r0;
	reg [7:0] index_register;
	reg [4:0] mapper1_shift_register = 0;

	// VRC4 prescaler
	reg [6:0] vrc4_irq_prescaler = 0;
	reg [1:0] vrc4_irq_prescaler_cycle = 0;
	reg vrc4_irq_prescaler_overflow = 0;

	// for CPU-based interrupts
	reg [15:0] irq_cpu_counter = 0;
	reg [15:0] irq_cpu_latch = 0;
	reg irq_cpu_counter_enable = 0;
	reg irq_cpu_counter_enable_after_ack = 0;
	reg irq_cpu_counter_cycle_mode = 0;
	reg irq_cpu_counter_16bit_mode = 0;
	reg irq_cpu_counter_reload = 0;
	reg irq_cpu_enable = 1'b1;
	reg irq_cpu_out = 0;

	// for scanline-based interrupts
	reg [2:0] a12_low_time;
	reg [7:0] irq_scanline_counter;
	reg [7:0] irq_scanline_counter_save;
	reg irq_scanline_counter_zero_save;
	reg irq_scanline_reload_save;
	reg irq_scanline_reload = 0;
	reg [7:0] irq_scanline_latch = 0;
	reg irq_scanline_enabled = 0;
	reg irq_scanline_out = 0;

	// for MMC2/MMC4
	reg ppu_latch0 = 0;
	reg ppu_latch1 = 0;

	reg readed;

	wire cpu_data_out_enabled;
	wire [7:0] cpu_data_out;
	assign cpu_data_out_enabled =
		(m2 & romsel & cpu_rw_in & ~lockout & (cpu_addr_in[14:12] == 3'b101)) ? 1'b1 : 1'b0;
//		(m2 & romsel & cpu_rw_in & (mapper == 5'b00000) & (cpu_addr_in[14:12] == 3'b101)) ? 1'b1 : 1'b0;
	assign cpu_data_out = 8'b00000000;
		

	assign ppu_ciram_a10 = (USE_MAPPER_118 & (mapper == 5'b10100) & flags[0]) ? ppu_addr_mapped[17] :
		(mirroring[1] ? mirroring[0] : (mirroring[0] ? ppu_addr_in[11] : ppu_addr_in[10])); // vertical / horizontal, 1Sa, 1Sb			

	wire [20:13] cpu_addr_mapped = // (map_rom_on_6000 & romsel & m2) ? prg_bank_6000 :
	(
		prg_mode[2] ? (
			prg_mode[1] ? (				
				prg_mode[0] ? (
					// 111 - 0x8000(A)
					{prg_bank_a[7:2], cpu_addr_in[14:13]}
				) : (
					// 110 - 0x8000(B)
					{prg_bank_b[7:2], cpu_addr_in[14:13]}
				)
			) : ( // prg_mode[1]
				prg_mode[0] ? (
					// 101 - 0x2000(C)+0x2000(B)+0x2000(A)+0x2000(D)
					cpu_addr_in[14] ? (cpu_addr_in[13] ? prg_bank_d : prg_bank_a) : (cpu_addr_in[13] ? prg_bank_b : prg_bank_c)
				) : ( // prg_mode[0]
					// 100 - 0x2000(A)+0x2000(B)+0x2000(C)+0x2000(D)
					cpu_addr_in[14] ? (cpu_addr_in[13] ? prg_bank_d : prg_bank_c) : (cpu_addr_in[13] ? prg_bank_b : prg_bank_a)
				)
			)
		) : ( // prg_mode[2]
			prg_mode[0] ? (
				// 0x1 - 0x4000(C) + 0x4000 (A)
				{cpu_addr_in[14] ? prg_bank_a[7:1] : prg_bank_c[7:1], cpu_addr_in[13]}
			) : ( // prg_mode[0]
				// 0x0 - 0x4000(A) + 0x4000 (С)
				{cpu_addr_in[14] ? prg_bank_c[7:1] : prg_bank_a[7:1], cpu_addr_in[13]}
			)
		)
	);

	wire [17:10] ppu_addr_mapped = (USE_MAPPER_021_022_023_025 && USE_MAPPER_022 && (mapper == 5'b11000) & flags[1]) ?
		{1'b1, ppu_addr_base[17:11]} : ppu_addr_base;

	wire [17:10] ppu_addr_base = chr_mode[2] ? (
		chr_mode[1] ? (
			chr_mode[0] ? ( 
				// 111 - 0x400(A)+0x400(B)+0x400(C)+0x400(D)+0x400(E)+0x400(F)+0x400(G)+0x400(H)
				ppu_addr_in[12] ? 
					(ppu_addr_in[11] ? (ppu_addr_in[10] ? chr_bank_h : chr_bank_g) : 
					(ppu_addr_in[10] ? chr_bank_f : chr_bank_e)) : (ppu_addr_in[11] ? (ppu_addr_in[10] ? chr_bank_d : chr_bank_c) : (ppu_addr_in[10] ? chr_bank_b : chr_bank_a))
			) : ( // chr_mode[0]
				// 110 - 0x800(A)+0x800(C)+0x800(E)+0x800(G)
				{ppu_addr_in[12] ? 
					(ppu_addr_in[11] ? chr_bank_g[7:1] : chr_bank_e[7:1]) : 
					(ppu_addr_in[11] ? chr_bank_c[7:1] : chr_bank_a[7:1]), ppu_addr_in[10]}
			)
		) : ( // chr_mode[1]
			// 100 - 0x1000(A) + 0x1000(E)
			// 101 - 0x1000(A/B) + 0x1000(E/F) - MMC2 и MMC4
			{ppu_addr_in[12] ?
				(((USE_MAPPER_009_010) && chr_mode[0] && ppu_latch1) ? chr_bank_f[7:2] : chr_bank_e[7:2]) : 
				(((USE_MAPPER_009_010) && chr_mode[0] && ppu_latch0) ? chr_bank_b[7:2] : chr_bank_a[7:2]),
			ppu_addr_in[11:10]}
		)
	) : ( // chr_mode[2]
		chr_mode[1] ? (
			// 010 - 0x800(A)+0x800(C)+0x400(E)+0x400(F)+0x400(G)+0x400(H) 
			// 011 - 0x400(E)+0x400(F)+0x400(G)+0x400(H)+0x800(A)+0x800(С)
			(ppu_addr_in[12]^chr_mode[0]) ? 
				(ppu_addr_in[11] ?
					(ppu_addr_in[10] ? chr_bank_h : chr_bank_g) : 
					(ppu_addr_in[10] ? chr_bank_f : chr_bank_e)
				) : (
					ppu_addr_in[11] ? {chr_bank_c[7:1],ppu_addr_in[10]} : {chr_bank_a[7:1],ppu_addr_in[10]}
				)
		) : ( // chr_mode[1]			
				// 000 - 0x2000(A)
				{chr_bank_a[7:3], ppu_addr_in[12:10]}
			 )
	);

	assign irq = (irq_scanline_out | irq_cpu_out) ? 1'b0 : 1'bZ;

	/* ------------------------------------------------------------------------------------------------- */

	// for VRC
	wire vrc_2b_hi = cpu_addr_in_low[1] | cpu_addr_in_low[3]; // | cpu_addr_in[5] | cpu_addr_in[7];
	wire vrc_2b_low = cpu_addr_in_low[0] | cpu_addr_in_low[2]; // | cpu_addr_in[4] | cpu_addr_in[6];

	/* ------------------------------------------------------------------------------------------------- */

	wire int_clk;
	IntOSC IntOSC_inst (.oscena(1'b1), .osc(int_clk));

	// watchdog counter run on internal oscillator
	always @ (posedge int_clk)
	begin
		if (~m2)
		begin
			reset <= 0;
			wdt_counter <= 0;
		end else
		begin
			// some black magic for fitting in 24 LABs
			{reset, wdt_counter} <= {wdt_counter[2] & lockout, wdt_counter[1] & lockout, wdt_counter[0] & lockout} + 1'b1;
		end
	end

	/* ------------------------------------------------------------------------------------------------- */

	// gated write clock
	always @ (posedge m2)
	begin
		cpu_rw_in_store <= cpu_rw_in;
	end

	always @ (*)
	begin
		write_clock = cpu_rw_in_store | m2;
	end

	always @ (negedge m2)
	begin
		// block two writes in a row (RMW) for games like Snow Bros. and Bill & Ted's Excellent Adventure
		// also you can remove this check and just patch those games, lol
		if (DOUBLE_WRITE_FIX)
			readed <= cpu_rw_in;
	end

	/* ------------------------------------------------------------------------------------------------- */

	// VRC4 IRQ prescaler (113,114,114 cycles)
	always @ (posedge m2)
	begin
		if (USE_MAPPER_021_022_023_025 && USE_VRC4_INTERRUPTS)
		begin
			if (irq_cpu_counter_enable)
			begin
				if (vrc4_irq_prescaler_overflow || irq_cpu_counter_reload)
				begin
					vrc4_irq_prescaler_overflow <= 0;

					if (vrc4_irq_prescaler_cycle[1] || irq_cpu_counter_reload)
					begin
						vrc4_irq_prescaler <= 7'b0010000; // 128-113=15(0x0f) +1 cycle
						vrc4_irq_prescaler_cycle <= 2'b00; // clear on reset or 3'rd cycle
					end else
					begin
						vrc4_irq_prescaler <= 7'b0001111; // 128-114=14(0x0e) +1 cycle
						vrc4_irq_prescaler_cycle <= {vrc4_irq_prescaler_cycle[0], 1'b1}; // shift register
					end
				end else
					{vrc4_irq_prescaler_overflow, vrc4_irq_prescaler} <= vrc4_irq_prescaler + 1'b1;
			end
		end
	end

	/* ------------------------------------------------------------------------------------------------- */

	always @ (negedge m2, posedge reset)
	begin
		if (reset)
		begin
			if (USE_MAPPER_021_022_023_025 && USE_VRC4_INTERRUPTS) // 8 bits
				irq_cpu_counter[7:0] <= 0;
			if (USE_MAPPER_069 || USE_MAPPER_073) // 8/16 bits
				irq_cpu_counter <= 0;

			if ((USE_MAPPER_021_022_023_025 && USE_VRC4_INTERRUPTS) || USE_MAPPER_073)
				irq_cpu_counter_reload <= 1'b0;

			if ((USE_MAPPER_021_022_023_025 && USE_VRC4_INTERRUPTS) || USE_MAPPER_069 || USE_MAPPER_073)
				irq_cpu_out <= 1'b0;
		end else
		begin
			if (irq_cpu_counter_enable)
			begin
				if (USE_MAPPER_021_022_023_025 && USE_VRC4_INTERRUPTS)
				begin
					if (irq_cpu_counter_cycle_mode || vrc4_irq_prescaler_overflow || irq_cpu_counter_reload)
					begin
						if ((irq_cpu_counter[7:0] == 8'hff) || irq_cpu_counter_reload)
							irq_cpu_counter <= irq_cpu_latch;
						else
							irq_cpu_counter[7:0] <= irq_cpu_counter[7:0] + 1'b1;

						if (irq_cpu_counter[7:0] == 8'hff)
							irq_cpu_out <= 1'b1;
					end
				end

				if (USE_MAPPER_073)
				begin
					if (irq_cpu_counter_16bit_mode)
					begin
						if ((irq_cpu_counter == 16'hffff) || irq_cpu_counter_reload)
							irq_cpu_counter <= irq_cpu_latch;
						else
							irq_cpu_counter <= irq_cpu_counter + 1'b1;

						if (irq_cpu_counter == 16'hffff)
							irq_cpu_out <= 1'b1;
					end else
					begin
						if ((irq_cpu_counter[7:0] == 8'hff) || irq_cpu_counter_reload)
							irq_cpu_counter <= irq_cpu_latch;
						else
							irq_cpu_counter[7:0] <= irq_cpu_counter[7:0] + 1'b1;

						if (irq_cpu_counter[7:0] == 8'hff)
							irq_cpu_out <= 1'b1;
					end
				end

				if (irq_cpu_counter_reload)
					irq_cpu_counter_reload <= 1'b0;

				if (USE_MAPPER_069)
				begin
					irq_cpu_counter <= irq_cpu_counter - 1'b1;
					if (irq_cpu_counter == 16'h0000)
						irq_cpu_out <= irq_cpu_enable;
				end
			end

			// moved from main block because of different clocks
			if (!cpu_rw_in_store)
			begin
				if (!romsel)
				begin
					if (USE_MAPPER_073 && (mapper == 5'b10011))
					begin
						case (cpu_addr_in[14:12])
							3'b100: begin // $C000-$CFFF
									irq_cpu_counter_reload <= cpu_data_in[1];
									irq_cpu_out <= 1'b0;
								end
							3'b101: irq_cpu_out <= 1'b0; // $D000-$DFFF
						endcase
					end

					if (USE_MAPPER_021_022_023_025  && USE_VRC4_INTERRUPTS && (mapper == 5'b11000))
					begin
						case ({cpu_addr_in[14:12], flags[0] ? vrc_2b_low : vrc_2b_hi, flags[0] ? vrc_2b_hi : vrc_2b_low})
							5'b11110: begin // $F002
											irq_cpu_counter_reload <= cpu_data_in[1];
											irq_cpu_out <= 1'b0;
										end
							5'b11111: irq_cpu_out <= 1'b0; // $F003
						endcase
					end

					if (USE_MAPPER_069 && (mapper == 5'b11001))
					begin
						if (cpu_addr_in[14:13] == 2'b01)
						begin
							case (index_register[3:0])
								4'b1101: irq_cpu_out <= 1'b0;
								4'b1110: irq_cpu_counter[7:0] <= cpu_data_in; // IRQ low
								4'b1111: irq_cpu_counter[15:8] <= cpu_data_in; // IRQ high
							endcase
						end
					end
				end // romsel
			end // write
		end
	end

	/* ------------------------------------------------------------------------------------------------- */

	always @ (negedge write_clock, posedge reset)
	begin
		if (reset)
		begin
			cpu_base <= 0;
			prg_mask <= 7'b1111000;
			chr_mask <= 0;
			prg_mode <= 0;
			chr_mode <= 0;
			prg_bank_a[5:1] <= 0;
			chr_bank_a[7:3] <= 0;
			mapper <= 0;
			flags <= 0;
			sram_enabled <= 0;
			prg_write_enabled <= 0;
			mirroring <= 0;
			lockout <= 0;

			if (USE_MAPPER_001)
			begin
				mapper1_shift_register <= 0;
				prg_bank_c[5:0] <= 6'b111110;
				chr_bank_a[6:2] <= 0;
			end
			if (USE_MAPPER_004)
			begin
				index_register[2:0] <= 0;
				prg_bank_a[(MMC3_BITSIZE-1):0] <= 0;
				prg_bank_b[(MMC3_BITSIZE-1):0] <= 0;
				chr_bank_a <= 0;
				chr_bank_c <= 0;
				chr_bank_e <= 0;
				chr_bank_f <= 0;
				chr_bank_g <= 0;
				chr_bank_h <= 0;
			end
			if (USE_MAPPER_021_022_023_025)
			begin
				prg_bank_a[4:0] <= 0;
				prg_bank_b[4:0] <= 0;
				chr_bank_a <= 0;
				chr_bank_b <= 0;
				chr_bank_c <= 0;
				chr_bank_d <= 0;
				chr_bank_e <= 0;
				chr_bank_f <= 0;
				chr_bank_g <= 0;
				chr_bank_h <= 0;
				irq_cpu_counter_cycle_mode <= 0;
			end
			if (USE_MAPPER_033_048)
			begin
				prg_bank_a[5:0] <= 0;
				prg_bank_b[5:0] <= 0;
				chr_bank_a <= 0;
				chr_bank_c <= 0;
				chr_bank_e <= 0;
				chr_bank_f <= 0;
				chr_bank_g <= 0;
				chr_bank_h <= 0;
			end
			if (USE_MAPPER_069)
			begin
				prg_bank_a[5:0] <= 0;
				prg_bank_b[5:0] <= 0;
				prg_bank_c[5:0] <= 6'b111110;
				chr_bank_a <= 0;
				chr_bank_b <= 0;
				chr_bank_c <= 0;
				chr_bank_d <= 0;
				chr_bank_e <= 0;
				chr_bank_f <= 0;
				chr_bank_g <= 0;
				chr_bank_h <= 0;
				irq_cpu_counter_enable <= 0;
				irq_cpu_enable <= 1'b1;
			end
			if (USE_MAPPER_073)
			begin
				prg_bank_a[3:1] <= 0;
				irq_cpu_counter_16bit_mode <= 0;
			end

			if (USE_MAPPER_090)
			begin
				prg_bank_a[5:0] <= 0;
				prg_bank_b[5:0] <= 0;
				prg_bank_c[5:0] <= 6'b111110;
				prg_bank_d[5:0] <= 6'b111111;
				chr_bank_a <= 0;
				chr_bank_b <= 0;
				chr_bank_c <= 0;
				chr_bank_d <= 0;
				chr_bank_e <= 0;
				chr_bank_f <= 0;
				chr_bank_g <= 0;
				chr_bank_h <= 0;
			end

			if ((USE_MAPPER_021_022_023_025 && USE_VRC4_INTERRUPTS) || USE_MAPPER_073)
			begin
				irq_cpu_counter_enable <= 0;
				irq_cpu_counter_enable_after_ack <= 0;
			end

			if (USE_MAPPER_004 || (USE_MAPPER_033_048 && USE_MAPPER_048_INTERRUPTS) || USE_MAPPER_090)
			begin
				irq_scanline_reload <= 0;
				irq_scanline_latch <= 0;
				irq_scanline_enabled <= 0;
			end
		end else
		begin
			if (romsel) // $0000-$7FFF
			begin
				if ((cpu_addr_in[14:12] == 3'b101) && ~lockout) // $5000-5FFF & lockout is off
				begin
					case (cpu_addr_in_low[2:0]) // case (cpu_addr_in[2:0])
						3'b000: // $5xx0
							{cpu_base[26:22]} <= cpu_data_in[4:0];						// CPU base address A26-A22
						3'b001: // $5xx1
							cpu_base[21:14] <= cpu_data_in[7:0]; 						// CPU base address A21-A14
						3'b010: // $5xx2
							prg_mask[20:14] <= cpu_data_in[6:0]; 						// CPU mask A18-A14
						3'b011: // $5xx3
							{prg_mode[2:0], chr_bank_a[7:3]} <= cpu_data_in[7:0];	// PRG mode, direct chr_bank_a access
						3'b100: // $5xx4
							{chr_mode[2:0], chr_mask[17:13]} <= cpu_data_in[7:0];	// CHR mode, CHR mask A17-A13
						3'b101: // $5xx5
							//{prg_bank_a[5:1], sram_page[1:0]} = cpu_data_in[6:0]; // direct prg_bank_a access, current SRAM page 0-3
							{prg_bank_a[5:1]} <= cpu_data_in[6:2]; // direct prg_bank_a access
						3'b110: // $5xx6
							{flags[2:0], mapper[4:0]} <= cpu_data_in[7:0];			// some flags, mapper
						3'b111: // $5xx7														
							// some other parameters						
							//{lockout, four_screen, mirroring[1:0], prg_write_enabled, chr_write_enabled, sram_enabled} = {cpu_data_in[7], cpu_data_in[5:0]};
							{lockout, mirroring[1:0], prg_write_enabled, sram_enabled} <= {cpu_data_in[7], cpu_data_in[4:2], cpu_data_in[0]};
					endcase

					if (USE_MAPPER_009_010 && mapper == 5'b10001) prg_bank_b <= 8'b11111101;
				end

				// Mapper #189
				// It's MMC3 with flag1
				if (USE_MAPPER_189 && flags[1] && (mapper == 5'b10100))
				begin
					//if (cpu_addr_in[14:0] >= 15'h4120) // $4120-$7FFF
					if (cpu_addr_in[14:8] >= 7'h41) // $4120-$7FFF
					begin
						prg_bank_a[5:2] <= cpu_data_in[3:0] | cpu_data_in[7:4];
					end
				end				

			end else begin // $8000-$FFFF

				// Mapper #2 - UxROM
				// flag0 - mapper #71 - for Fire Hawk only.
				// other mapper-#71 games are UxROM
				if (USE_MAPPER_002 && (mapper == 5'b00001))
				begin
					if (!USE_MAPPER_071 || ~flags[0] || (cpu_addr_in[14:12] != 3'b001))
					begin
						prg_bank_a[5:1] <= cpu_data_in[4:0];
					end else begin // CodeMasters, blah. Mirroring control used only by Fire Hawk
						mirroring[1:0] <= {1'b1, cpu_data_in[4]};
					end
				end
				
				// Mapper #3 - CNROM
				if (USE_MAPPER_003 && (mapper == 5'b00010))
				begin
					chr_bank_a[7:3] <= cpu_data_in[4:0];
				end

				// Mapper #7 - AxROM, mapper #241 - BNROM
				if (USE_MAPPER_007 && (mapper == 5'b01000))
				begin
					prg_bank_a[5:2] <= cpu_data_in[3:0];
					if (!USE_MAPPER_241 || !flags[0]) // BNROM?
						mirroring <= {1'b1, cpu_data_in[4]};
				end

				// Mapper #11 - ColorDreams
				if (USE_MAPPER_011 && (mapper == 5'b01010))
				begin
					prg_bank_a[3:2] <= cpu_data_in[1:0];
					chr_bank_a[6:3] <= cpu_data_in[7:4];
				end

				// Mapper #66 - GxROM
				if (USE_MAPPER_066 && (mapper == 5'b01011))
				begin
					prg_bank_a[3:2] <= cpu_data_in[5:4];
					chr_bank_a[4:3] <= cpu_data_in[1:0];					
				end

				// Mapper #90 - JY
				if (USE_MAPPER_090 && (mapper == 5'b01101))
				begin
					if (cpu_addr_in[14:12] == 3'b000) // $800x
					begin
						case (cpu_addr_in_low[1:0])
							2'b00: prg_bank_a[5:0] <= cpu_data_in[5:0];
							2'b01: prg_bank_b[5:0] <= cpu_data_in[5:0];
							2'b10: prg_bank_c[5:0] <= cpu_data_in[5:0];
							2'b11: prg_bank_d[5:0] <= cpu_data_in[5:0];
						endcase
					end
					if (cpu_addr_in[14:12] == 3'b001) // $900x
					begin
						case (cpu_addr_in_low[2:0])
							3'b000: chr_bank_a <= cpu_data_in;
							3'b001: chr_bank_b <= cpu_data_in;
							3'b010: chr_bank_c <= cpu_data_in;
							3'b011: chr_bank_d <= cpu_data_in;
							3'b100: chr_bank_e <= cpu_data_in;
							3'b101: chr_bank_f <= cpu_data_in;
							3'b110: chr_bank_g <= cpu_data_in;
							3'b111: chr_bank_h <= cpu_data_in;
						endcase
					end
					if ({cpu_addr_in[14:12], cpu_addr_in_low[1:0]} == 5'b10101) // $D001
						mirroring <= cpu_data_in[1:0];
					if (/*USE_MAPPER_090_INTERRUPTS &&*/ cpu_addr_in[14:12] == 3'b100) // $C00x
					begin
						if (cpu_addr_in_low[2:0] == 3'b000)
							irq_scanline_enabled <= cpu_data_in[0];
						if (cpu_addr_in_low[2:1] == 2'b01)
							irq_scanline_enabled <= cpu_addr_in_low[0];
						if (cpu_addr_in_low[2:0] == 3'b101)
						begin
							irq_scanline_latch <= cpu_data_in ^ r0;
							irq_scanline_reload <= ~irq_scanline_reload;
						end
						if (cpu_addr_in_low[2:0] == 3'b110)
							r0 <= cpu_data_in;
					end
				end
				
				// Mapper #1 - MMC1
				/*
				flag0 - 16KB of WRAM (SOROM)
				*/
				if (USE_MAPPER_001 && (mapper == 5'b10000))
				begin
					if (!DOUBLE_WRITE_FIX || readed)
					begin
						mapper1_shift_register <= (cpu_data_in[7] | mapper1_shift_register[0]) ? 5'b10000
							: {cpu_data_in[0], mapper1_shift_register[4:1]};
					end

					if (mapper1_shift_register[0]) // last bit in cpu_data_in[0] for sync mode
					begin
						case (cpu_addr_in[14:13])
							2'b00: begin // $8000-$9FFF
								prg_mode <= mapper1_shift_register[4] ? {2'b00, ~mapper1_shift_register[3]} : 3'b111;
								prg_bank_c[4:0] <= mapper1_shift_register[3] ? 5'b11110 : 5'b00000;
								chr_mode <= {cpu_data_in[0], 2'b00};
								mirroring <= {~mapper1_shift_register[2], mapper1_shift_register[1]};
							end
							2'b01: begin // $A000-$BFFF
								chr_bank_a[6:2] <= {cpu_data_in[0], mapper1_shift_register[4:1]};
								prg_bank_a[5] <= cpu_data_in[0]; // for SUROM, 512k PRG support
								prg_bank_c[5] <= cpu_data_in[0]; // for SUROM, 512k PRG support
							end
							2'b10: chr_bank_e[6:2] <= {cpu_data_in[0], mapper1_shift_register[4:1]}; // $C000-$DFFF
							2'b11: begin // $E000-$FFFF
								prg_bank_a[4:1] <= mapper1_shift_register[4:1]; // $E000-$FFFF
								sram_enabled <= ~cpu_data_in[0];
							end
						endcase
					end

					if (cpu_data_in[7]) // reset
					begin
						prg_mode <= 3'b000;	// 0x4000 (A) + fixed last (C)
						prg_bank_c[4:0] <= 5'b11110;
					end
				end
				
				// Mapper #9 and #10 - MMC2 and MMC4
				// flag0 - 0=MMC2, 1=MMC4
				if (USE_MAPPER_009_010 && (mapper == 5'b10001))
				begin
					case (cpu_addr_in[14:12])
						3'b010: if (~flags[0]) // $A000-$AFFF
										prg_bank_a[3:0] <= cpu_data_in[3:0];
									else
										prg_bank_a[4:1] <= cpu_data_in[3:0];
						3'b011: chr_bank_a[6:2] <= cpu_data_in[4:0]; // $B000-$BFFF
						3'b100: chr_bank_b[6:2] <= cpu_data_in[4:0]; // $C000-$CFFF
						3'b101: chr_bank_e[6:2] <= cpu_data_in[4:0]; // $D000-$DFFF
						3'b110: chr_bank_f[6:2] <= cpu_data_in[4:0]; // $E000-$EFFF
						3'b111: mirroring <= {1'b0, cpu_data_in[0]}; // $F000-$FFFF
					endcase
				end

				// Mapper #152
				if (USE_MAPPER_152 && (mapper == 5'b10010))
				begin
					chr_bank_a[6:3] <= cpu_data_in[3:0];
					prg_bank_a[3:1] <= cpu_data_in[6:4];
					mirroring <= {1'b1, cpu_data_in[7]};
				end		

				// Mapper #73 - VRC3
				if (USE_MAPPER_073 && (mapper == 5'b10011))
				begin
					case (cpu_addr_in[14:12])
						3'b000: irq_cpu_latch[3:0] <= cpu_data_in[3:0]; // $8000-$8FFF
						3'b001: irq_cpu_latch[7:4] <= cpu_data_in[3:0]; // $9000-$9FFF
						3'b010: irq_cpu_latch[11:8] <= cpu_data_in[3:0]; // $A000-$AFFF
						3'b011: irq_cpu_latch[15:12] <= cpu_data_in[3:0]; // $B000-$BFFF
						3'b100: begin // $C000-$CFFF
								irq_cpu_counter_16bit_mode <= ~cpu_data_in[2];
								irq_cpu_counter_enable <= cpu_data_in[1];
								irq_cpu_counter_enable_after_ack <= cpu_data_in[0];
							end
						3'b101: begin // $D000-$DFFF
								irq_cpu_counter_enable <= irq_cpu_counter_enable_after_ack;
							end
						3'b110: ; // $E000-$EFFF
						3'b111: prg_bank_a[3:1] <= cpu_data_in[2:0]; // $F000-$FFFF
					endcase
				end			

				// Mapper #4 - MMC3/MMC6
				/*
				r0[2:0] - internal register
				flag0 - TxSROM
				flag1 - mapper #189
				*/
				if (USE_MAPPER_004 && (mapper == 5'b10100))
				begin
					case ({cpu_addr_in[14:13], cpu_addr_in_low[0]})
						3'b000: begin // $8000-$9FFE, even
							index_register[2:0] <= cpu_data_in[2:0];
							chr_mode <= {2'b01, cpu_data_in[7]};
							if (!USE_MAPPER_189 || ~flags[1])
								prg_mode <= {2'b10, cpu_data_in[6]};
						end
						3'b001: begin // $8001-$9FFF, odd
							case (index_register[2:0])
								3'b000: chr_bank_a <= cpu_data_in;
								3'b001: chr_bank_c <= cpu_data_in;
								3'b010: chr_bank_e <= cpu_data_in;
								3'b011: chr_bank_f <= cpu_data_in;
								3'b100: chr_bank_g <= cpu_data_in;
								3'b101: chr_bank_h <= cpu_data_in;
								3'b110: if (!USE_MAPPER_189 || ~flags[1]) prg_bank_a[(MMC3_BITSIZE-1):0] <= cpu_data_in[(MMC3_BITSIZE-1):0];
								3'b111: if (!USE_MAPPER_189 || ~flags[1]) prg_bank_b[(MMC3_BITSIZE-1):0] <= cpu_data_in[(MMC3_BITSIZE-1):0];
							endcase
						end
						3'b010: mirroring <= {1'b0, cpu_data_in[0]}; // $A000-$BFFE, even (mirroring)
						3'b100: irq_scanline_latch <= cpu_data_in; // $C000-$DFFE, even (IRQ latch)
						3'b101: irq_scanline_reload <= ~irq_scanline_reload; // $C001-$DFFF, odd
						//3'b110, irq_scanline_enabled <= 0; // $E000-$FFFE, even
						//3'b111: irq_scanline_enabled <= 1; // $E001-$FFFF, odd						
					endcase
					if (cpu_addr_in[14] && cpu_addr_in[13])
						irq_scanline_enabled <= cpu_addr_in_low[0];
				end

				// Mappers #33 + #48 - Taito
				// flag0=0 - #33, flag0=1 - #48
				if (USE_MAPPER_033_048 && (mapper == 5'b10110))
				begin
					case ({cpu_addr_in[14:13], cpu_addr_in_low[1:0]})
						4'b0000: begin
							prg_bank_a[5:0] <= cpu_data_in[5:0]; // $8000, PRG Reg 0 (8k @ $8000)
							if (~flags[0]) // #33
								mirroring <= {1'b0, cpu_data_in[6]};
						end
						4'b0001: prg_bank_b[5:0] <= cpu_data_in[5:0]; // $8001, PRG Reg 1 (8k @ $A000)
						4'b0010: chr_bank_a <= {cpu_data_in[6:0], 1'b0};  // $8002, CHR Reg 0 (2k @ $0000)
						4'b0011: chr_bank_c <= {cpu_data_in[6:0], 1'b0};  // $8003, CHR Reg 1 (2k @ $0800)
						4'b0100: chr_bank_e <= cpu_data_in;	// $A000, CHR Reg 2 (1k @ $1000)
						4'b0101: chr_bank_f <= cpu_data_in; // $A001, CHR Reg 2 (1k @ $1400)
						4'b0110: chr_bank_g <= cpu_data_in; // $A002, CHR Reg 2 (1k @ $1800)
						4'b0111: chr_bank_h <= cpu_data_in; // $A003, CHR Reg 2 (1k @ $1C00)
						4'b1100: if (flags[0]) mirroring <= {1'b0, cpu_data_in[6]};	// $E000, mirroring, for mapper #48
					endcase

					if (USE_MAPPER_048_INTERRUPTS)
					begin
						case ({cpu_addr_in[14:13], cpu_addr_in_low[1:0]})
							4'b1000: irq_scanline_latch <= cpu_data_in; // $C000, IRQ latch
							4'b1001: irq_scanline_reload <= ~irq_scanline_reload; // $C001, IRQ reload
							4'b1010: irq_scanline_enabled <= 1; // $C002, IRQ enable
							4'b1011: irq_scanline_enabled <= 0; // $C003, IRQ disable & ack
						endcase
					end
				end

				// Mapper #23 - VRC2/4
				/*
				flag0 - switches A0 and A1 lines. 0=A0,A1 like VRC2b (mapper #23), 1=A1,A0 like VRC2a(#22), VRC2c(#25)
				flag1 - divides CHR bank select by two (mapper #22, VRC2a)
				*/
				if (USE_MAPPER_021_022_023_025 && mapper == 5'b11000)
				begin
					case ({cpu_addr_in[14:12], flags[0] ? vrc_2b_low : vrc_2b_hi, flags[0] ? vrc_2b_hi : vrc_2b_low})
						5'b00100,
						5'b00101: if (cpu_data_in != 8'b11111111) mirroring <= cpu_data_in[1:0]; // $9000-$9001, mirroring
						5'b00110,
						5'b00111: prg_mode[0] <= cpu_data_in[1]; // $9002-$9004, PRG swap
						5'b01100: chr_bank_a[3:0] <= cpu_data_in[3:0];  // $B000, CHR0 low
						5'b01101: chr_bank_a[7:4] <= cpu_data_in[3:0];  // $B001, CHR0 hi
						5'b01110: chr_bank_b[3:0] <= cpu_data_in[3:0];  // $B002, CHR1 low
						5'b01111: chr_bank_b[7:4] <= cpu_data_in[3:0];  // $B003, CHR1 hi
						5'b10000: chr_bank_c[3:0] <= cpu_data_in[3:0];  // $C000, CHR2 low
						5'b10001: chr_bank_c[7:4] <= cpu_data_in[3:0];  // $C001, CHR2 hi
						5'b10010: chr_bank_d[3:0] <= cpu_data_in[3:0];  // $C002, CHR3 low
						5'b10011: chr_bank_d[7:4] <= cpu_data_in[3:0];  // $C003, CHR3 hi
						5'b10100: chr_bank_e[3:0] <= cpu_data_in[3:0];  // $D000, CHR4 low
						5'b10101: chr_bank_e[7:4] <= cpu_data_in[3:0];  // $D001, CHR4 hi
						5'b10110: chr_bank_f[3:0] <= cpu_data_in[3:0];  // $D002, CHR5 low
						5'b10111: chr_bank_f[7:4] <= cpu_data_in[3:0];  // $D003, CHR5 hi
						5'b11000: chr_bank_g[3:0] <= cpu_data_in[3:0];  // $E000, CHR6 low
						5'b11001: chr_bank_g[7:4] <= cpu_data_in[3:0];  // $E001, CHR6 hi
						5'b11010: chr_bank_h[3:0] <= cpu_data_in[3:0];  // $E002, CHR7 low
						5'b11011: chr_bank_h[7:4] <= cpu_data_in[3:0];  // $E003, CHR7 hi
					endcase
					if (cpu_addr_in[14:12] == 3'b000)
						prg_bank_a[4:0] <= cpu_data_in[4:0];  // $8000-$8003, PRG0
					if (cpu_addr_in[14:12] == 3'b010)
						prg_bank_b[4:0] <= cpu_data_in[4:0];  // $A000-$A003, PRG1
						
					if (USE_VRC4_INTERRUPTS)
					begin
						case ({cpu_addr_in[14:12], flags[0] ? vrc_2b_low : vrc_2b_hi, flags[0] ? vrc_2b_hi : vrc_2b_low})
							5'b11100: irq_cpu_latch[3:0] <= cpu_data_in[3:0];  // IRQ latch low
							5'b11101: irq_cpu_latch[7:4] <= cpu_data_in[3:0];  // IRQ latch hi
							5'b11110: begin // IRQ control
											irq_cpu_counter_cycle_mode <= cpu_data_in[2];
											irq_cpu_counter_enable <= cpu_data_in[1];
											irq_cpu_counter_enable_after_ack <= cpu_data_in[0];
										end
							5'b11111: begin // IRQ ack
											irq_cpu_counter_enable <= irq_cpu_counter_enable_after_ack;
										end
						endcase
					end
				end

				// Mapper #69 - Sunsoft FME-7
				if (USE_MAPPER_069 && mapper == 5'b11001)
				begin
					if (cpu_addr_in[14:13] == 2'b00)
						index_register[3:0] <= cpu_data_in[3:0];

					if (cpu_addr_in[14:13] == 2'b01)
					begin
						case (index_register[3:0])
							4'b0000: chr_bank_a <= cpu_data_in; // CHR0
							4'b0001: chr_bank_b <= cpu_data_in; // CHR1
							4'b0010: chr_bank_c <= cpu_data_in; // CHR2
							4'b0011: chr_bank_d <= cpu_data_in; // CHR3
							4'b0100: chr_bank_e <= cpu_data_in; // CHR4
							4'b0101: chr_bank_f <= cpu_data_in; // CHR5
							4'b0110: chr_bank_g <= cpu_data_in; // CHR6
							4'b0111: chr_bank_h <= cpu_data_in; // CHR7
							4'b1000: sram_enabled <= cpu_data_in[7]; // PRG0
							4'b1001: prg_bank_a[5:0] <= cpu_data_in[5:0]; // PRG1
							4'b1010: prg_bank_b[5:0] <= cpu_data_in[5:0]; // PRG2
							4'b1011: prg_bank_c[5:0] <= cpu_data_in[5:0]; // PRG3
							4'b1100: mirroring[1:0] <= cpu_data_in[1:0]; // mirroring
							4'b1101: begin
											irq_cpu_counter_enable <= cpu_data_in[7];
											irq_cpu_enable <= cpu_data_in[0];
										end
						endcase
					end						
				end
			end // romsel
		end // write
	end

	/* ------------------------------------------------------------------------------------------------- */

	// A12 must be low for 3 rises of M2
	always @ (posedge m2, posedge ppu_addr_in[12])
	begin
		if (ppu_addr_in[12])
			a12_low_time <= 0;
		else
			a12_low_time <= {a12_low_time[1:0], 1'b1}; // shift register
	end

	// IRQ counter
	always @ (posedge m2)
	begin
		if (ppu_rd_in) // only if PPU not reading to avoid glitches
		begin
			if ((irq_scanline_reload ^ irq_scanline_reload_save) || irq_scanline_counter_zero_save)
				irq_scanline_counter <= irq_scanline_latch;
			else
				// count down for MMC3 and up for Taito
				irq_scanline_counter <= irq_scanline_counter_save +
					((USE_MAPPER_033_048 && USE_MAPPER_048_INTERRUPTS && (mapper == 5'b10110)) ? 1'b1 : -1'b1);
		end
	end

	// only copy prepared values on actual interrupt
	always @ (posedge ppu_addr_in[12])
	begin
		if (a12_low_time[2])
		begin
			irq_scanline_reload_save <= irq_scanline_reload;
			irq_scanline_counter_save <= irq_scanline_counter;
			irq_scanline_counter_zero_save <= irq_scanline_counter == 8'h00;
		end
	end

	// IRQ line control
	always @ (posedge ppu_addr_in[12], negedge irq_scanline_enabled)
	begin
		if (~irq_scanline_enabled)
			irq_scanline_out <= 1'b0;
		else
		begin
			if (a12_low_time[2])
				irq_scanline_out <= irq_scanline_counter == 8'h00;
		end
	end

	/* ------------------------------------------------------------------------------------------------- */


	// for MMC2/MMC4
	always @ (negedge ppu_rd_in)
	begin
		if (USE_MAPPER_009_010)
		begin
			if (ppu_addr_in[13:3] == 11'b00111111011) ppu_latch0 <= 0;
			if (ppu_addr_in[13:3] == 11'b00111111101) ppu_latch0 <= 1;
			if (ppu_addr_in[13:3] == 11'b01111111011) ppu_latch1 <= 0;
			if (ppu_addr_in[13:3] == 11'b01111111101) ppu_latch1 <= 1;
		end
	end

endmodule
